/*
 * =============================================================================
 * Simplified BSD License, see http://www.opensource.org/licenses/
 * -----------------------------------------------------------------------------
 * Copyright (c) 2008-2009, Marco Terzer, Zurich, Switzerland
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice, 
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright 
 *       notice, this list of conditions and the following disclaimer in the 
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Swiss Federal Institute of Technology Zurich 
 *       nor the names of its contributors may be used to endorse or promote 
 *       products derived from this software without specific prior written 
 *       permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 * POSSIBILITY OF SUCH DAMAGE.
 * =============================================================================
 */
package ch.javasoft.metabolic.efm.column;

import ch.javasoft.bitset.IBitSet;

/**
 * A pair of {@link Column} instances. May be used to encapsulate adjacency
 * candidates. A column pair instance is constant, meaning that no member 
 * variables change after instantiation.
 */
public class ColumnPair<Col extends Column> {

	private final Col colA, colB;
	
	public ColumnPair(Col colA, Col colB) {
		this.colA = colA;
		this.colB = colB;
	}
	public Col getColumnA() {
		return colA;
	}

	public Col getColumnB() {
		return colB;
	}
	
	/**
	 * Returns the intersection of the bit values.
	 */
	public final IBitSet intersectBitValues() {
		return colA.bitValues().getAnd(colB.bitValues());
	}

	public boolean isSuperSet(Col superCandidate) {
		return 
			!superCandidate.equals(colA) && !superCandidate.equals(colB) &&
			colA.bitValues().isSubSetOf(superCandidate.bitValues());
	}
	public boolean existsSuperSet(Iterable<Col> superCandidates) {
		IBitSet inter = null;
		for (Col cand : superCandidates) {
			if (!cand.equals(colA) && !cand.equals(colB)) {				
				if (inter == null) {
					inter = intersectBitValues(); 
				}
				if (inter.isSubSetOf(cand.bitValues())) {
					return true;
				}
			}
		}
		return false;
	}
	
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("{a=");
		sb.append(getColumnA());
		sb.append(", b=");
		sb.append(getColumnB());
		sb.append(", a^b=");
		sb.append(intersectBitValues());
		sb.append("}");
		return sb.toString();
	}

}
